/*
   PN532 NFC module routines for STM32F4xx-nucleo

     Copyright (C) by Y.Ishioka
*/
#include "stm32f4xx.h"
#include "spi.h"
#include "nfc.h"
#include <stdio.h>
#include <string.h>

void  delay( int  msec );
void  WaitUsec( int usec );
void  syslog( unsigned int prio, const char *format, ... );


/*
 * return  0: invalid
 *         1: accept
 */
int  nfc_read_irq( void )
{
  int  stat = 0;
  unsigned int  data;

  data = GPIOA->IDR ;
  
  if( (data & 0x0100) != 0 ) {
    stat = 1;
  }

  return  stat;
}


/*************************************************/
#define _BV(bit) (1<<(bit))

#define PN532_PACKBUFFSIZ   64
uint8_t pn532_packetbuffer[ PN532_PACKBUFFSIZ ];

unsigned char  pn532ack[] = {0x00, 0x00, 0xFF, 0x00, 0xFF, 0x00};
unsigned char   pn532response_firmwarevers[] = {0x00, 0xFF, 0x06, 0xFA, 0xD5, 0x03};


/*************************************************/
inline  void  spi_write( uint8_t c )
{
  uint8_t tmp = 0x00;

  if( c & 0x01 ) {
    tmp |= 0x80;
  }
  if( c & 0x02 ) {
    tmp |= 0x40;
  }
  if( c & 0x04 ) {
    tmp |= 0x20;
  }
  if( c & 0x08 ) {
    tmp |= 0x10;
  }
  if( c & 0x10 ) {
    tmp |= 0x08;
  }
  if( c & 0x20 ) {
    tmp |= 0x04;
  }
  if( c & 0x40 ) {
    tmp |= 0x02;
  }
  if( c & 0x80 ) {
    tmp |= 0x01;
  }

  xspi_send( (uint16_t)tmp );
}


/*************************************************/
inline uint8_t  spi_read( void )
{
  uint8_t  tmp = 0x00;
  uint8_t  c;

  c = xspi_recv();

  if( c & 0x01 ) {
    tmp |= 0x80;
  }
  if( c & 0x02 ) {
    tmp |= 0x40;
  }
  if( c & 0x04 ) {
    tmp |= 0x20;
  }
  if( c & 0x08 ) {
    tmp |= 0x10;
  }
  if( c & 0x10 ) {
    tmp |= 0x08;
  }
  if( c & 0x20 ) {
    tmp |= 0x04;
  }
  if( c & 0x40 ) {
    tmp |= 0x02;
  }
  if( c & 0x80 ) {
    tmp |= 0x01;
  }

  return  (uint8_t)tmp;
}; 


/*************************************************/
int  isready( void )
{
    // SPI read status and check if ready.
    xspi_set_cs_nfc(0);
    delay(2);
    spi_write(PN532_SPI_STATREAD);

    // read byte
    uint8_t x = spi_read();

    xspi_set_cs_nfc(1);

    // Check if status is ready.
    return  x == PN532_SPI_READY;
}


/*************************************************/
int  waitready( uint16_t timeout )
{
  uint16_t timer = 0;
  while(!isready()) {
    if (timeout != 0) {
      timer += 10;
      if (timer > timeout) {
        //syslog( 5, "TIMEOUT!" );
        return  0;
      }
    }
    delay(10);
  }
  return  1;
}


/*************************************************/
void  readdata( uint8_t* buff, uint8_t n )
{
    // SPI write.
    xspi_set_cs_nfc(0);
    delay(2);
    spi_write(PN532_SPI_DATAREAD);

    for (uint8_t i=0; i<n; i++) {
      //delay(1);
      buff[i] = spi_read();
    }

    xspi_set_cs_nfc(1);
}


/*************************************************/
void  writecommand( uint8_t* cmd, uint8_t cmdlen )
{
    // SPI command write.
    uint8_t checksum;

    cmdlen++;

    xspi_set_cs_nfc(0);
    delay(2);     // or whatever the delay is for waking up the board
    spi_write(PN532_SPI_DATAWRITE);

    checksum = PN532_PREAMBLE + PN532_PREAMBLE + PN532_STARTCODE2;
    spi_write(PN532_PREAMBLE);
    spi_write(PN532_PREAMBLE);
    spi_write(PN532_STARTCODE2);

    spi_write(cmdlen);
    spi_write(~cmdlen + 1);

    spi_write(PN532_HOSTTOPN532);
    checksum += PN532_HOSTTOPN532;

    for (uint8_t i=0; i<cmdlen-1; i++) {
      spi_write(cmd[i]);
      checksum += cmd[i];
    }

    spi_write(~checksum);
    spi_write(PN532_POSTAMBLE);
    xspi_set_cs_nfc(1);
}


/*************************************************/
int  readack( void )
{
  uint8_t ackbuff[6];

  readdata(ackbuff, 6);

  return (0 == strncmp((char *)ackbuff, (char *)pn532ack, 6));
}


/*************************************************/
int  sendCommandCheckAck(uint8_t *cmd, uint8_t cmdlen, uint16_t timeout)
{
  // write the command
  writecommand(cmd, cmdlen);

  // Wait for chip to say its ready!
  if (!waitready(timeout)) {
    return  0;
  }

  //while( nfc_read_irq() == 0 ) ;

  // read acknowledgement
  if (!readack()) {
    return  0;
  }

  if (!waitready(timeout)) {
    return  0;
  }

  return  1; // ack'd command
}


/*************************************************/
uint32_t getFirmwareVersion(void)
{
  uint32_t response;

  pn532_packetbuffer[0] = PN532_COMMAND_GETFIRMWAREVERSION;

  if (! sendCommandCheckAck(pn532_packetbuffer, 1, 100)) {
    return 0;
  }

  // read data packet
  readdata(pn532_packetbuffer, 12);

  // check some basic stuff
  if (0 != strncmp((char *)pn532_packetbuffer, (char *)pn532response_firmwarevers, 6)) {
    return 0;
  }

#if 0
  syslog( 5, "0x%02x 0x%02x 0x%02x",
           pn532_packetbuffer[0], pn532_packetbuffer[1], pn532_packetbuffer[2] );
  syslog( 5, "0x%02x 0x%02x 0x%02x",
           pn532_packetbuffer[3], pn532_packetbuffer[4], pn532_packetbuffer[5] );
  syslog( 5, "0x%02x 0x%02x 0x%02x 0x%02x",
           pn532_packetbuffer[6], pn532_packetbuffer[7], pn532_packetbuffer[8], pn532_packetbuffer[9] );
  syslog( 5, "0x%02x 0x%02x 0x%02x 0x%02x",
           pn532_packetbuffer[10], pn532_packetbuffer[11], pn532_packetbuffer[12], pn532_packetbuffer[13] );
#endif

  int offset = 7;  // Skip a response byte
  response = pn532_packetbuffer[offset++];
  response <<= 8;
  response |= pn532_packetbuffer[offset++];
  response <<= 8;
  response |= pn532_packetbuffer[offset++];
  response <<= 8;
  response |= pn532_packetbuffer[offset++];

  return response;
}


/*************************************************/
int  SAMConfig( void )
{
  pn532_packetbuffer[0] = PN532_COMMAND_SAMCONFIGURATION;
  pn532_packetbuffer[1] = 0x01; // normal mode;
  pn532_packetbuffer[2] = 0x14; // timeout 50ms * 20 = 1 second
  pn532_packetbuffer[3] = 0x01; // use IRQ pin!

  if (! sendCommandCheckAck(pn532_packetbuffer, 4, 100))
    return  0;

  // read data packet
  readdata(pn532_packetbuffer, 8);

  int offset = 5;
  return  (pn532_packetbuffer[offset] == 0x15);
}


/*************************************************/
int  setPassiveActivationRetries( uint8_t maxRetries )
{
  pn532_packetbuffer[0] = PN532_COMMAND_RFCONFIGURATION;
  pn532_packetbuffer[1] = 5;    // Config item 5 (MaxRetries)
  pn532_packetbuffer[2] = 0xFF; // MxRtyATR (default = 0xFF)
  pn532_packetbuffer[3] = 0x01; // MxRtyPSL (default = 0x01)
  pn532_packetbuffer[4] = maxRetries;

  if (! sendCommandCheckAck(pn532_packetbuffer, 5, 100))
    return 0x0;  // no ACK

  return 1;
}


/*************************************************/
int  readPassiveTargetID(uint8_t cardbaudrate, uint8_t * uid, uint8_t * uidLength, uint16_t timeout)
{
  pn532_packetbuffer[0] = PN532_COMMAND_INLISTPASSIVETARGET;
  pn532_packetbuffer[1] = 1;  // max 1 cards at once (we can set this to 2 later)
  pn532_packetbuffer[2] = cardbaudrate;

  if (!sendCommandCheckAck(pn532_packetbuffer, 3, timeout))
  {
    return 0x0;  // no cards read
  }

  // wait for a card to enter the field (only possible with I2C)
    if (!waitready(timeout)) {
      return 0x0;
    }

  // read data packet
  readdata(pn532_packetbuffer, 20);
  // check some basic stuff

  /* ISO14443A card response should be in the following format:

    byte            Description
    -------------   ------------------------------------------
    b0..6           Frame header and preamble
    b7              Tags Found
    b8              Tag Number (only one used in this example)
    b9..10          SENS_RES
    b11             SEL_RES
    b12             NFCID Length
    b13..NFCIDLen   NFCID                                      */

  if (pn532_packetbuffer[7] != 1)
    return 0;

  uint16_t sens_res = pn532_packetbuffer[9];
  sens_res <<= 8;
  sens_res |= pn532_packetbuffer[10];

  /* Card appears to be Mifare Classic */
  *uidLength = pn532_packetbuffer[12];
  for (uint8_t i=0; i < pn532_packetbuffer[12]; i++)
  {
    uid[i] = pn532_packetbuffer[13+i];
  }

  return 1;
}


/*************************************************/
void  debug_nfc_test( void )
{
  int success;
  uint8_t uid[7];    // Buffer to store the returned UID
  uint8_t uidLength; // Length of the UID (4 or 7 bytes depending on ISO14443A card type)

  success = readPassiveTargetID( PN532_MIFARE_ISO14443A, &uid[0], &uidLength, 100 );
  if( success != 0 ) {
    syslog( 5, "uidLen=%d", uidLength );
    syslog( 5, "uid=0x%02x 0x%02x 0x%02x 0x%02x", uid[0], uid[1], uid[2], uid[3] );
    syslog( 5, "    0x%02x 0x%02x 0x%02x", uid[4], uid[5], uid[6] );
  }
}


/*************************************************/
/*
 * success : 1 : success
 *           0 : fail
 */
int  nfc_get_uid( unsigned char *uid, unsigned char *uidLength )
{
  int success;

  success = readPassiveTargetID( PN532_MIFARE_ISO14443A, uid, uidLength, 100 );

  return  success;
}


/*************************************************/
void  nfc_dev_init( void )
{
    /* set clock enable */
    RCC->AHB1ENR |= (  RCC_AHB1ENR_GPIOAEN
                     | RCC_AHB1ENR_GPIOBEN
                    ) ;

    /** for NFC_IRQ:PA8 **/
    GPIOA->MODER   &= ~(0x00030000);
    GPIOA->OTYPER  &= ~(0x0100);
    GPIOA->OSPEEDR &= ~(0x00030000);
    GPIOA->OSPEEDR |=  (0x00020000);
    GPIOA->PUPDR   &= ~(0x00030000);
    //GPIOA->PUPDR   |=  (0x00010000); /* pull-up */

    /** for NFC_RST:PB10 **/
    GPIOB->MODER   &= ~(0x00300000);
    GPIOB->MODER   |=  (0x00100000);
    GPIOB->OTYPER  &= ~(0x0400);
    GPIOB->OSPEEDR &= ~(0x00300000);
    GPIOB->OSPEEDR |=  (0x00200000);
    GPIOB->PUPDR   &= ~(0x00300000);
    GPIOB->PUPDR   |=  (0x00100000); /* pull-up */

    GPIOB->BSRR = 0x0400 << 16;      /* NFC_RST:0 */
    delay(1);
    GPIOB->BSRR = 0x0400 ;           /* NFC_RST:1 */
    delay(1);
}


/*************************************************/
void  nfc_init( void )
{
  unsigned int  version;

  nfc_dev_init();

  // not exactly sure why but we have to send a dummy command to get synced up
  pn532_packetbuffer[0] = PN532_COMMAND_GETFIRMWAREVERSION;
  sendCommandCheckAck(pn532_packetbuffer, 1, 100);

  version = getFirmwareVersion();
  syslog( 5, "version=0x%08x", version );

  syslog( 5, "setPassiveActivationRetries()" );
  setPassiveActivationRetries(0xFF);
  syslog( 5, "SAMConfig()" );
  SAMConfig();

#if 0
  debug_nfc_test();
#endif

}
